"""Object that defines the structure of a FIT file message field."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import collections

from .data import Data, Schema
from .base_type import BaseType


class FieldDefinitionBase(Data, BaseType):
    """Base object that defines the structure of a FIT file message field."""

    base_type = None
    size = None

    def invalid(self):
        """Return the invalid value for the field."""
        return self._invalid(self.base_type)

    def type_string(self):
        """Return the type string for the field."""
        return self._type_string(self.base_type)

    def type_count(self):
        """Return the number of values for the field."""
        type_size = Schema.type_to_size[self.type_string()]
        return int(self.size / type_size)


class FieldDefinition(FieldDefinitionBase):
    """Object that defines the structure of a FIT file message field."""

    field_definition_number = None
    size = None
    base_type = None

    fd_schema = Schema(
        'FieldDefinition',
        collections.OrderedDict(
            [
                ('field_definition_number', ['UINT8', 1]),
                ('size', ['UINT8', 1]),
                ('base_type', ['UINT8', 1])
            ]
        )
    )

    def __init__(self, file):
        """
        Return a FieldDefinition instance created by reading data from a FIT file.

        Paramters:
        ---------
            file (File):  a FIT File instance.
        """
        super().__init__(file, FieldDefinition.fd_schema)

    def __str__(self):
        """Return a string representation for the FieldDefinition instance."""
        return f'FieldDefinition: type {self.field_definition_number}: {self.type_count()} of {self.type_string()}'



"""Objects that represent FIT file personal record message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import Field, TimeMsField, PowerField, UnknownField, CaloriesField, TimeMinField
from .enum_fields import EnumField
from .object_fields import DistanceCentimetersToMetersField, AltitudeField
from .field_enums import PersonalRecordType, GoalType, GoalRecurrence, GoalSource


class PersonalRecordTypeField(EnumField):
    """Field that indicates what type of personal record the matching PersonalRecordField field holds."""

    _name = 'pr_type'
    _enum = PersonalRecordType


class PersonalRecordField(Field):
    """Field that holds a personal record."""

    _name = 'personal_record'
    _dependant_field_control_fields = ['pr_type']

    _type_to_fields = {
        PersonalRecordType.time         : TimeMsField,
        PersonalRecordType.distance     : DistanceCentimetersToMetersField,
        PersonalRecordType.elevation    : AltitudeField,
        PersonalRecordType.power        : PowerField
    }

    def dependant_field(self, control_value_list):
        """Return a field class that should be used to handle a dependant field."""
        pr_type = control_value_list[0]
        field_name = 'unknown_pr'
        if pr_type is not None:
            try:
                _dependant_field = self._type_to_fields[pr_type]
                field_name = pr_type.name
            except Exception:
                _dependant_field = UnknownField
        else:
            _dependant_field = Field
        return _dependant_field(field_name)


class GoalTypeField(EnumField):
    """Field that indicates what type of goal the matching GoalValueField field holds."""

    _name = 'goal_type'
    _enum = GoalType


class GoalRecurrenceField(EnumField):
    """Field that indicates how frequently the matching goal happens."""

    _name = 'goal_recurrence'
    _enum = GoalRecurrence


class GoalSourceField(EnumField):
    """A class that holds a field indicating what the source of a goal is as a enumeration."""

    _name = 'goal_source'
    _enum = GoalSource


class GoalValueField(Field):
    """Field that holds a goal value."""

    _name = 'target_value'
    _dependant_field_control_fields = ['type']

    _type_to_fields = {
        GoalType.time              : TimeMsField,
        GoalType.distance          : DistanceCentimetersToMetersField,
        GoalType.calories          : CaloriesField,
        GoalType.frequency         : Field,
        GoalType.steps             : Field,
        GoalType.ascent            : AltitudeField,
        GoalType.active_minutes    : TimeMinField
    }

    def dependant_field(self, control_value_list):
        """Return the dependant field class for this instance."""
        goal_type = control_value_list[0]
        field_name = 'unknown_goal'
        if goal_type is not None:
            try:
                _dependant_field = self._type_to_fields[goal_type]
                field_name = goal_type.name
            except Exception:
                _dependant_field = UnknownField
        else:
            _dependant_field = Field
        return _dependant_field(field_name)



"""Test FIT file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

from fitfile import field_enums
from fitfile.product import GarminProduct
from fitfile import manufacturer_product_fields as mp_fields


root_logger = logging.getLogger()
handler = logging.FileHandler('fit_field_enum.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestFitDependantField(unittest.TestCase):
    """Class for testing FIT file parsing."""

    @classmethod
    def setUpClass(cls):
        pass

    def test_product_field_reconvert(self):
        field_value = mp_fields.ProductField().convert(GarminProduct.Fenix_5_Sapphire.value, GarminProduct.invalid.value)[0]
        self.assertEqual(field_value['product'], GarminProduct.Fenix_5_Sapphire.value)
        field_value.field = mp_fields.GarminProductField()
        field_value.reconvert(field_enums.DisplayMeasure.metric)
        self.assertIsInstance(field_value['product'], mp_fields.GarminProduct)
        self.assertEqual(field_value['product'], GarminProduct.Fenix_5_Sapphire)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3

"""Test measurement parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

from fitfile import Distance, Weight, Temperature, Speed


root_logger = logging.getLogger()
handler = logging.FileHandler('measurment.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestMeasurement(unittest.TestCase):
    """Class for testing measurment parsing."""

    @classmethod
    def setUpClass(cls):
        pass

    def test_distance(self):
        distance = Distance.from_mm(1000000)
        self.assertEqual(distance.value, 1000.0)
        self.assertEqual(distance.to_mm(), 1000000.0)
        self.assertEqual(distance.to_meters(), 1000.0)
        self.assertEqual(distance.to_kms(), 1.0)
        self.assertEqual(distance.to_mm(), 1000000)
        self.assertEqual(distance.to_miles(), 0.6213712)

    def test_distance_from_func(self):
        self.assertEqual(Distance.from_func('km'), Distance.from_kilometers)
        self.assertEqual(Distance.from_func('Kilometers'), Distance.from_kilometers)
        self.assertEqual(Distance.from_func('Feet'), Distance.from_feet)
        self.assertEqual(Distance.from_func('MeterS'), Distance.from_meters)

    def test_distance_from_func_raises(self):
        self.assertRaises(Exception, Distance.from_func, 'xxx')

    def test_speed_from_func(self):
        self.assertEqual(Speed.from_func('KPH'), Speed.from_kph)
        self.assertEqual(Speed.from_func('Mph'), Speed.from_mph)

    def test_speed_from_func_raises(self):
        self.assertRaises(Exception, Speed.from_func, 'xxx')

    def test_weight(self):
        weight = Weight.from_grams(100000)
        self.assertEqual(weight.value, 100.0)
        self.assertAlmostEqual(weight.to_lbs(), 220.46, 2)

    def test_temperature(self):
        temp = Temperature.from_celsius(100.0)
        self.assertEqual(temp.value, 100.0)
        self.assertEqual(temp.to_f(), 212.0)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Library for parsing FIT files and returning them as lists of messages dictionaries."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .file import File
from .message_type import UnknownMessageType, MessageType
from .file_type import FileType
from .manufacturer import Manufacturer
from .product import GarminProduct, GarminLocalProduct, ScoscheProduct, WahooFitnessProduct, UnknownProduct, product_enum
from .sport import Sport, SubSport
from .device_enums import MainDeviceType, AntplusDeviceType, LocalDeviceType, UnknownDeviceType
from .units import UnitTypes, unit_strings
from .measurement import Distance, Latitude, Longitude, Speed, Weight, Volume, Temperature, Cadence
from .version_info import version_string, __version__



"""Objects that represent FIT file object message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import Field
from .enum_fields import EnumField
from .manufacturer import Manufacturer
from .product import GarminProduct, GarminLocalProduct, ScoscheProduct, WahooFitnessProduct, HealthAndLifeProduct, UnknownProduct


class ManufacturerField(EnumField):
    """A field indicating the manufacturer of the device used to create the FIT file."""

    _name = 'manufacturer'
    _enum = Manufacturer

    def _convert_single(self, value, invalid):
        try:
            return self._enum(value)
        except Exception:
            if value >= Manufacturer.Garmin_local_start.value:
                return Manufacturer.Garmin_local
            return Manufacturer.unknown


class BaseProductField(EnumField):
    """A field indicating the product id of the device used to create the FIT file."""

    _name = 'product'


class GarminProductField(BaseProductField):
    """A field indicating the Garmin product id of the device used to create the FIT file."""

    _enum = GarminProduct


class GarminLocalProductField(BaseProductField):
    """A field indicating the Garmin local product id of the device used to create the FIT file."""

    _enum = GarminLocalProduct


class WahooFitnessProductField(BaseProductField):
    """A field indicating the Wahoo product id of the device used to create the FIT file."""

    _enum = WahooFitnessProduct


class ScoscheProductField(BaseProductField):
    """A field indicating the Scosche product id of the device used to create the FIT file."""

    _enum = ScoscheProduct


class HealthAndLifeProductField(BaseProductField):
    """A field indicating the Scosche product id of the device used to create the FIT file."""

    _enum = HealthAndLifeProduct


class UnknownProductField(BaseProductField):
    """A field indicating the undocumented product id of the device used to create the FIT file."""

    _enum = UnknownProduct


class ProductField(Field):
    """A field indicating the product id of the device used to create the FIT file."""

    _name = 'product'
    _dependant_field_control_fields = ['manufacturer']

    _manufacturer_to_product_fields = {
        Manufacturer.Garmin                 : GarminProductField,
        Manufacturer.Dynastream             : GarminProductField,
        Manufacturer.Dynastream_OEM         : GarminProductField,
        Manufacturer.Wahoo_Fitness          : WahooFitnessProductField,
        Manufacturer.Scosche                : ScoscheProductField,
        Manufacturer.Health_and_Life        : HealthAndLifeProductField,
        Manufacturer.Garmin_local           : GarminLocalProductField,
        Manufacturer.Garmin_local_154       : GarminLocalProductField,
        Manufacturer.Garmin_local_218       : GarminLocalProductField,
        Manufacturer.unknown                : UnknownProductField,
        Manufacturer.invalid                : GarminProductField,
    }

    def dependant_field(self, control_value_list):
        """Return a dependant field instance given the control field values."""
        manufacturer = control_value_list[0]
        try:
            dependant_field_name = self._manufacturer_to_product_fields[manufacturer]
        except Exception:
            dependant_field_name = UnknownProductField
        return dependant_field_name()



"""Objects that represent FIT file device message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import Field
from .type_fields import BitField, FloatField
from .enum_fields import EnumField
from .field_enums import BatteryStatus, BacklightMode, Side, DisplayOrientation, AntNetwork, SourceType, AutoSyncFrequency, BodyLocation, WatchFaceMode
from .device_enums import MainDeviceType, AntplusDeviceType, LocalDeviceType, UnknownDeviceType


class BatteryVoltageField(FloatField):
    """A Field that holds a battery voltage reading for the device."""

    _name = 'battery_voltage'
    _units = 'v'
    _scale = 256.0


class BatteryStatusField(EnumField):
    """A Field that holds a battery status reading of the device."""

    _name = 'battery_status'
    _enum = BatteryStatus


class AutoActivityDetectField(BitField):
    """A filed that contains a mask of activitiy types that will automatically be tracked."""

    _name = 'auto_activity_detect'
    _bits = {
        0x00000000: 'none',
        0x00000001: 'running',
        0x00000002: 'cycling',
        0x00000004: 'swimming',
        0x00000008: 'walking',
        0x00000020: 'elliptical',
        0x00000400: 'sedentary',
        0xffffffff: 'invalid'
    }


class DisplayOrientationField(EnumField):
    """A Field that holds the display orientation setting for the device."""

    _name = 'display_orientation'
    _enum = DisplayOrientation


class SideField(EnumField):
    """A Field that holds the left/right wrist setting for the device."""

    _name = 'side'
    _enum = Side


class BacklightModeField(EnumField):
    """A Field that holds the backlight setting for the device."""

    _name = 'backlight_mode'
    _enum = BacklightMode


class AntNetworkField(EnumField):
    """A field that holds an Ant network type."""

    _name = 'ant_network'
    _enum = AntNetwork


class SourceTypeField(EnumField):
    """A field that holds the source type of device."""

    _name = 'source_type'
    _enum = SourceType


class MainDeviceTypeField(EnumField):
    """Device types for devices connected to the main device via ANT+. Like a external heart rate monitor."""

    _name = 'main_device_type'
    _manufacturer = None
    _product = None
    _enum = MainDeviceType

    def is_invalid(self, value, invalid):
        """Return if the field's value is valid."""
        return False

    def _convert_single(self, value, invalid):
        """Return a device type for the device inferred from its manufactuer and product information."""
        return MainDeviceType.derive_device_type(self._manufacturer, self._product)


class AntplusDeviceTypeField(EnumField):
    """Device types for devices connected to the main device via ANT+. Like a external heart rate monitor."""

    _name = 'antplus_device_type'
    _enum = AntplusDeviceType


class LocalDeviceTypeField(EnumField):
    """Device types for sub-devices resident on the main device. Like sensors embedded on a watch."""

    _name = 'local_device_type'
    _enum = LocalDeviceType


class UnknownDeviceTypeField(EnumField):
    """Device type enums for device types that we don't know how to interpret."""

    _name = 'unknown_device_type'
    _enum = UnknownDeviceType


class DeviceTypeField(Field):
    """Wrapper field for device type dependant fields."""

    _name = 'device_type'
    _dependant_field_control_fields = ['source_type', 'device_type', 'manufacturer', 'product']

    _source_to_device_type_fields = {
        SourceType.ant: Field,
        SourceType.antplus: AntplusDeviceTypeField,
        SourceType.local: LocalDeviceTypeField,
    }

    def dependant_field(self, control_value_list):
        """Return a field class that should be used to handle a dependant field."""
        source_type = control_value_list[0]
        device_type = control_value_list[1]
        manufacturer = control_value_list[2]
        product = control_value_list[3]
        if source_type is not None:
            if source_type is SourceType.local and device_type is None:
                dependant_field = MainDeviceTypeField(manufacturer=manufacturer, product=product, name='device_type')
            else:
                try:
                    dependant_field = self._source_to_device_type_fields[source_type](name='device_type')
                except Exception:
                    dependant_field = UnknownDeviceTypeField(name='device_type')
        else:
            dependant_field = Field(name='device_type')
        return dependant_field


class AutoSyncFrequencyField(EnumField):
    """A filed that holds the maxiimum amount of time before the device automatically syncs with the cloud."""

    _name = 'auto_sync_frequency'
    _enum = AutoSyncFrequency


class BodyLocationField(EnumField):
    """A field that identifies a location of a device on the body."""

    _name = 'body_location'
    _enum = BodyLocation


class WatchFaceModeField(EnumField):
    """A field that identifies the mode the watch face is in."""

    _name = 'watch_face_mode'
    _enum = WatchFaceMode



"""Library for parsing FIT files and returning them as lists of messages dictionaries."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .field_enums import UnknownEnumValue, FieldEnum, FuzzyFieldEnum, Switch, FitBaseUnit, DisplayMeasure, DisplayHeart, DisplayPosition, DisplayOrientation, Side, BacklightMode, \
    AntNetwork, SourceType, BatteryStatus, AutoSyncFrequency, BodyLocation, Gender, HeartRateZoneCalc, PowerCalc, Language, DateMode, TimeMode, Activity, ActivityType, Event, EventType, \
    LapTrigger, SessionTrigger, PersonalRecordType, GoalType, GoalRecurrence, GoalSource, WatchFaceMode, ClimbProEvent, HeartRateZonesTimerType, HeartRateZonesMethod, name_for_enum


"""Objects that represent FIT file device message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import Field, NamedField
from .field_definition import FieldDefinition


class FitBaseTypeField(NamedField):

    def _convert_single(self, value, invalid):
        if value != invalid:
            try:
                return FieldDefinition.type_name(value)
            except Exception:
                return value


class TypeField(NamedField):
    """A base class for fields based on python types."""

    @classmethod
    def __nop(cls, value):
        return value

    type_func = __nop

    def _convert_single(self, value, invalid):
        if value != invalid:
            try:
                return self.type_func(value / self._scale)
            except Exception:
                return value


class IntegerField(TypeField):
    """A FIT file message field with a integer value."""

    type_func = int


class FloatField(TypeField):
    """A FIT file message field with a float value."""

    type_func = float


class BoolField(TypeField):
    """A FIT file message field with a boolean value."""

    type_func = bool


class BitField(Field):
    """A FIT file message field with a bitfield value."""

    _bits = {}

    def _convert_single(self, value, invalid):
        if value != invalid:
            return self._bits.get(value, [self._bits[bit] for bit in self._bits if ((bit & value) == bit)])


class StringField(NamedField):
    """A FIT file message field with a string value."""

    def _invalid_single(self, value, invalid):
        return (value < 0) or (value > 127)

    def _convert_many(self, value, invalid):
        if isinstance(value, list):
            converted_value = ""
            for aschii_index in value:
                if aschii_index == 0:
                    break
                converted_value += chr(aschii_index)
        else:
            converted_value = str(value)
        return converted_value.strip()


class BytesField(NamedField):
    """A FIT file message field with a bytearray value."""

    def _convert_many(self, value, invalid):
        if isinstance(value, list):
            converted_value = bytearray()
            for character in value:
                converted_value.append(character)
        else:
            converted_value = bytearray(value)
        return converted_value


class HeartRateField(FloatField):
    """A FIT file message field holding a heart rate measurement."""

    _units = 'bpm'


class AbsolutePressureField(IntegerField):
    """Absolute pressure measure for diving?."""

    _name = 'absolute_pressure'
    _units = 'Pa'


class RespirationRateField(FloatField):
    """A respiration rate measurment in breaths per minute."""

    _name = 'respiration_rate'
    _units = 'brpm'
    _scale = 100.0


class Vo2MaxField(IntegerField):
    """A new Vo2 Max measurement."""

    _name = 'vo2_max'
    _units = 'L/min'
    _scale = 18724.5714


class MetaMaxField(IntegerField):
    """Reports metamax from which Vo2Max can be derived."""

    _name = 'metamax'
    _dependant_field_control_fields = []

    def dependant_field(self, control_value_list):
        """Return a Vo2MaxField as dependant field since it is derived from metamax."""
        return Vo2MaxField



"""FIT file data message."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import logging
import datetime

from .data_field import DataField
from .dev_data_field import DevDataField


logger = logging.getLogger(__name__)


class MessageFields(dict):
    """Class holds field names and values."""

    def __getattr__(self, name):
        return self.get(name)


class DataMessageDecodeContext():
    """Class that holds data used across decoding of all DataMessages."""

    def __init__(self):
        """Return a DataMessageDecodeContext instance."""
        self.matched_timestamp_16 = None
        self.last_timestamp = None
        self.last_absolute_timestamp = None

    def absolute_timestamp(self, absolute_timestamp):
        """Update the context given an absolute timestamp."""
        self.last_timestamp = absolute_timestamp
        self.last_absolute_timestamp = absolute_timestamp
        self.matched_timestamp_16 = None

    def timestamp16_to_timestamp(self, timestamp_16):
        """Calculate an absolute timestamp given a relative timestamp16."""
        if self.matched_timestamp_16:
            if timestamp_16 >= self.matched_timestamp_16:
                delta = timestamp_16 - self.matched_timestamp_16
            else:
                delta = (self.matched_timestamp_16 - 65535) + timestamp_16
        else:
            self.matched_timestamp_16 = timestamp_16
            delta = 0
        self.last_timestamp = self.last_absolute_timestamp + datetime.timedelta(seconds=delta)
        return self.last_timestamp


class DataMessage():
    """Class decodes and holds a FIT file data message."""

    def __init__(self, definition_message, fit_file, measurement_system, context):
        """Return a DataMessage instance decoded from the supplied FIT file using the supplied definition message."""
        self.__definition_message = definition_message
        self.__context = context
        self.fields = MessageFields()
        self.field_values = MessageFields()
        self.file_size = 0
        message_fields = {}
        for index in range(definition_message.fields):
            data_field = DataField(fit_file, definition_message, definition_message.field_definitions[index], measurement_system)
            self.file_size += data_field.file_size
            for field_value in data_field.values:
                message_fields[field_value.field.name] = field_value
        self.__convert_fields(message_fields, measurement_system)
        self.__convert_dev_fields(fit_file, measurement_system)
        self.__track_time()

    def __add_field(self, field_value):
        field_name = field_value.field.name.lower()
        self.fields.update(field_value)
        self.field_values[field_name] = field_value

    def __control_field_value(self, field, message_fields, control_field_name):
        if control_field_name in message_fields:
            return message_fields[control_field_name].first()

    def __convert_fields(self, message_fields, measurement_system):
        for field_value in message_fields.values():
            field = field_value.field
            if field._dependant_field_control_fields:
                control_values = [self.__control_field_value(field, message_fields, control_field) for control_field in field._dependant_field_control_fields]
                field_value.field = field.dependant_field(control_values)
                field_value.reconvert(measurement_system)
            self.__add_field(field_value)

    def __convert_dev_fields(self, fit_file, measurement_system):
        if self.__definition_message.has_dev_fields:
            for index in range(self.__definition_message.dev_fields):
                dev_data_field = DevDataField(fit_file, self.__definition_message, self.__definition_message.dev_field_definitions[index], measurement_system)
                self.file_size += dev_data_field.file_size
                self.__add_field(dev_data_field.value)

    def __track_time(self):
        if 'timestamp' in self.fields:
            self.__context.absolute_timestamp(self.fields.timestamp)
        elif 'timestamp_16' in self.fields:
            timestamp_16 = self.fields.timestamp_16
            if timestamp_16 is not None:
                self.fields['timestamp'] = self.__context.timestamp16_to_timestamp(timestamp_16)
            else:
                # This should not happen, if the timestamp16 field exists, it should not be None
                # Issue #21: seen on Ubuntu on Windows
                logger.error('timestamp16 with value None: %r', self.fields)
                self.fields.add('timestamp', self.__context.last_timestamp)

    @property
    def type(self):
        """Return the message type."""
        return self.__definition_message.message_type

    def __str__(self):
        """Return a string representation of a DataMessage instance."""
        return f'{self.__class__.__name__}: {repr(self.type)}: {str(self.fields)}'

    def __repr__(self):
        """Return a string representation of a DataMessage instance."""
        # we reformat the values with a list comprehension to avoid the .values() showing up as a generator in the repr output
        return f'{self.__class__.__name__}({repr(self.type)}: {repr(list(self.field_values.values()))})'



"""Objects that represents measurements."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .field_enums import DisplayMeasure


class Measurement():
    """Object that represents measurement."""

    def __init__(self, value, raw_value, invalid_value, default_output_func=None, default_output_units=None):
        """Return a Measurement instance."""
        self.value = value
        self.raw_value = raw_value
        self.invalid_value = invalid_value
        self.default_output_func = default_output_func
        self.default_output_units = default_output_units

    @classmethod
    def from_units(cls, value, scale, invalid_value=None):
        """Create a Distance object from a distance measurement."""
        return cls((value * scale) if value is not None else None, value, invalid_value)

    def to_units(self, scale=1.0, rounded=False):
        """Return a measurement value given the current value and a scale."""
        if not self.is_invalid():
            try:
                value = self.value * scale
                return round(value) if rounded else value
            except Exception as e:
                raise Exception(f'value {self.value} scale {scale}: {e}')

    def is_invalid(self):
        """Return if the measurement is valid."""
        return (self.raw_value is None) or (self.raw_value == self.invalid_value)

    def __eq__(self, other):
        if not isinstance(other, Measurement):
            return NotImplemented
        return (self.value == other.value)

    def __repr__(self):
        """Return a string representation of a Measurement instance."""
        if self.is_invalid():
            value = 'invalid'
        else:
            value = f'{self.default_output_func()} {self.default_output_units} [{self.raw_value}, {self.invalid_value}]'
        return f'{self.__class__.__name__ }({value})'

    def __str__(self):
        """Return a string representation of a Measurement instance."""
        return self.__repr__()


class Distance(Measurement):
    """An object that handles convertions of distance mesaurements from one representation to another."""

    def __init__(self, meters, raw_value, invalid_value):
        """Return an instance of the Distance class."""
        super().__init__(meters, raw_value, invalid_value, self.to_meters, 'meters')

    @classmethod
    def from_kilometers(cls, kms, invalid_value=None):
        """Create a Distance object from a distance measurement in kilometers."""
        return cls.from_units(kms, 1000.0, invalid_value)

    @classmethod
    def from_meters(cls, meters, invalid_value=None):
        """Create a Distance object from a distance measurement in meters."""
        return cls(meters, meters, invalid_value)

    @classmethod
    def from_cm(cls, cm, invalid_value=None):
        """Create a Distance object from a distance measurement in centimeters."""
        return cls.from_units(cm, 0.01, invalid_value)

    @classmethod
    def from_mm(cls, mm, invalid_value=None):
        """Create a Distance object from a distance measurement in millimeters."""
        return cls.from_units(mm, 0.001, invalid_value)

    @classmethod
    def from_feet(cls, feet, invalid_value=None):
        """Create a Distance object from a distance measurement in feet."""
        return cls.from_units(feet, 0.3048, invalid_value)

    @classmethod
    def from_miles(cls, feet, invalid_value=None):
        """Create a Distance object from a distance measurement in miles."""
        return cls.from_units(feet, 1609.344, invalid_value)

    @classmethod
    def from_func(cls, units):
        """Return a Distance object from function for a measurement in units."""
        units_to_obj_func = {
            'cm': cls.from_cm,
            'centimeters': cls.from_cm,
            'm': cls.from_meters,
            'meters': cls.from_meters,
            'km': cls.from_kilometers,
            'kilometers': cls.from_kilometers,
            'ft': cls.from_feet,
            'feet': cls.from_feet,
            'mi': cls.from_miles,
            'miles': cls.from_miles
        }
        try:
            return units_to_obj_func[units.lower()]
        except KeyError:
            raise Exception(f'No conversion for units {units}')

    @classmethod
    def from_meters_or_feet(cls, distance, measurement_system=DisplayMeasure.metric):
        """Return a distance object."""
        return cls.from_meters(distance) if measurement_system is DisplayMeasure.metric else cls.from_feet(distance)

    def to_mm(self):
        """Return the distance measurement as millimeters."""
        return self.to_units(1000.0)

    def to_meters(self):
        """Return the distance measurement as meters."""
        return self.to_units()

    def to_kms(self):
        """Return the distance measurement as kilometers."""
        return self.to_units(0.001)

    def to_inches(self):
        """Return the distance measurement as inches."""
        return self.to_units(39.37008)

    def to_feet(self):
        """Return the distance measurement as feet."""
        return self.to_units(3.2808399)

    def to_miles(self):
        """Return the distance measurement as miles."""
        return self.to_units(0.0006213712)

    @classmethod
    def inches_or_mm(cls, distance, measurement_system=DisplayMeasure.metric):
        """Return the distance measurement as millimeters or inches depending on the measurement system."""
        return distance.mm_or_inches(measurement_system)

    def mm_or_inches(self, measurement_system=DisplayMeasure.metric):
        """Return the distance measurement as millimeters or inches depending on the measurement system."""
        return self.to_mm() if measurement_system is DisplayMeasure.metric else self.to_inches()

    @classmethod
    def feet_or_meters(cls, distance, measurement_system=DisplayMeasure.metric):
        """Return the distance measurement as meters or feet depending on the measurement system."""
        return distance.meters_or_feet(measurement_system)

    def meters_or_feet(self, measurement_system=DisplayMeasure.metric):
        """Return the distance measurement as meters or feet depending on the measurement system."""
        return self.to_meters() if measurement_system is DisplayMeasure.metric else self.to_feet()

    #
    # @classmethod
    # def kms_or_miles(cls, distance, measurement_system=DisplayMeasure.metric):
    #     """Return the distance measurement as kilometers or miles depending on the measurement system."""
    #     return distance.kms_or_miles(measurement_system)

    def kms_or_miles(self, measurement_system=DisplayMeasure.metric):
        """Return the distance measurement as kilometers or miles depending on the measurement system."""
        return self.to_kms() if measurement_system is DisplayMeasure.metric else self.to_miles()


class Position(Measurement):
    """Object that represents position measurement."""

    def __init__(self, semicircles, raw_value, invalid_value):
        """Return an instance of the Position class."""
        super().__init__(semicircles, raw_value, invalid_value, self.to_degrees, 'degrees')

    @classmethod
    def from_semicircles(cls, semicircles, invalid_value=None):
        """Return a Position instance intialized from a value measured in semicircles."""
        return cls(semicircles, semicircles, invalid_value)

    # @classmethod
    # def to_degrees(cls, position, measurement_system=DisplayMeasure.metric):
    #     return position.to_degrees(measurement_system)

    def to_degrees(self, measurement_system=DisplayMeasure.metric):
        """Return the position measurement as degrees."""
        return self.to_units(180.0 / 2147483648.0)


class Latitude(Position):
    """Object that represents position latitude."""

    pass


class Longitude(Position):
    """Object that represents position longitude."""

    pass


class Speed(Measurement):
    """Object that represents a speed measurement."""

    def __init__(self, meters_per_sec, raw_value, invalid_value):
        """Return a Speed instance intialized with a value in meters per sec."""
        super().__init__(meters_per_sec, raw_value, invalid_value, self.to_mph, 'mph')

    @classmethod
    def from_mps(cls, meters_per_sec, invalid_value=None):
        """Return a Speed instance intialized with a value in meters per sec."""
        return cls(meters_per_sec, meters_per_sec, invalid_value)

    @classmethod
    def from_kph(cls, km_per_hour, invalid_value=None):
        """Return a Speed instance intialized with a value in kilometers per hour."""
        return cls.from_units(km_per_hour, 0.2777778, invalid_value)

    @classmethod
    def from_mph(cls, miles_per_hour, invalid_value=None):
        """Return a Speed instance intialized with a value in miles per hour."""
        return cls.from_units(miles_per_hour, 0.44704, invalid_value)

    @classmethod
    def from_cps(cls, centimeters_per_sec, invalid_value=None):
        """Return a Speed instance intialized with a value in centimeters per sec."""
        return cls.from_units(centimeters_per_sec, 0.01, invalid_value)

    @classmethod
    def from_mmps(cls, mm_per_sec, invalid_value=None):
        """Return a Speed instance intialized with a value in millimeters per sec."""
        return cls.from_units(mm_per_sec, 0.001, invalid_value)

    @classmethod
    def from_func(cls, units):
        """Return a Distance object from function for a measurement in units."""
        units_to_obj_func = {
            'mps'   : cls.from_mps,
            'km/h'  : cls.from_kph,
            'kph'   : cls.from_kph,
            'mph'   : cls.from_mph
        }
        try:
            return units_to_obj_func[units.lower()]
        except KeyError:
            raise Exception(f'No conversion for units {units}')

    @classmethod
    def from_kph_or_mph(cls, speed, measurement_system=DisplayMeasure.metric):
        """Return a Speed instance intialized with a value in kilometers per hopur or miles per hour."""
        return cls.from_kph(speed) if measurement_system is DisplayMeasure.metric else cls.from_mph(speed)

    def to_kph(self):
        """Return the speed measurement as kilometers per hour."""
        return self.to_units(3.6)

    def to_mph(self):
        """Return the speed measurement as miles per hour."""
        return self.to_units(2.236936)

    def to_mps(self):
        """Return the speed measurement as meters per second."""
        return self.value

    @classmethod
    def mph_or_kph(cls, speed, measurement_system=DisplayMeasure.metric):
        """Return the speed measurement as kilometers per hour or miles per hour."""
        return speed.kph_or_mph(measurement_system)

    def kph_or_mph(self, measurement_system=DisplayMeasure.metric):
        """Return the speed measurement as kilometers per hour or miles per hour."""
        return self.to_kph() if measurement_system is DisplayMeasure.metric else self.to_mph()


class Weight(Measurement):
    """Object that represents a weight measurement."""

    def __init__(self, kgs, raw_value, invalid_value):
        """Return a Weight instance intialized with a value in kilograms."""
        super().__init__(kgs, raw_value, invalid_value, self.to_lbs, 'lbs')

    @classmethod
    def from_grams(cls, grams, invalid_value=None):
        """Return a Weight instance created from a value in grams."""
        return cls.from_units(grams, 0.001, invalid_value)

    @classmethod
    def from_cgs(cls, centigrams, invalid_value=None):
        """Return a Weight instance created from a value in centigrams."""
        return cls.from_units(centigrams, 0.1, invalid_value)

    @classmethod
    def from_lbs(cls, lbs, invalid_value=None):
        """Return a Weight instance created from a value in pounds."""
        return cls.from_units(lbs, 0.4535924, invalid_value)

    def to_kgs(self):
        """Return the weight measurement as kilograms."""
        return self.value

    def to_lbs(self):
        """Return the weight measurement as pounds."""
        return self.to_units(2.204623)

    @classmethod
    def lbs_or_kgs(cls, weight, measurement_system=DisplayMeasure.metric):
        """Return the weight measurement as kilograms or pounds."""
        return weight.kgs_or_lbs(measurement_system)

    def kgs_or_lbs(self, measurement_system=DisplayMeasure.metric):
        """Return the weight measurement as kilograms or pounds."""
        return self.to_kgs() if measurement_system is DisplayMeasure.metric else self.to_lbs()


class Volume(Measurement):
    """Object that represents a volume measurement."""

    def __init__(self, kgs, raw_value, invalid_value):
        """Return a Volume instance intialized with a value in liters."""
        super().__init__(kgs, raw_value, invalid_value, self.to_ounces, 'Ounces')

    @classmethod
    def from_milliliters(cls, milliliters, invalid_value=None):
        """Return a Volume instance created from a value in milliliters."""
        return cls.from_units(milliliters, 0.001, invalid_value)

    def to_liters(self):
        """Return the volume measurement as liters."""
        return self.value

    def to_milliliters(self, rounded=False):
        """Return the volume measurement as milliliters."""
        return self.to_units(1000.0, rounded=rounded)

    def to_ounces(self, rounded=False):
        """Return the volume measurement as ounces."""
        return self.to_units(35.19503, rounded=rounded)

    def ml_or_oz(self, measurement_system=DisplayMeasure.metric, rounded=False):
        """Return the volume measurement as milliliters or ounces."""
        return self.to_milliliters(rounded) if measurement_system is DisplayMeasure.metric else self.to_ounces(rounded)


class Temperature(Measurement):
    """Object that represents a temperature measurement."""

    def __init__(self, celsius, raw_value, invalid_value):
        """Return a Temperature instance intialized with a value in celsius."""
        super().__init__(celsius, raw_value, invalid_value, self.to_f, 'F')

    @classmethod
    def from_celsius(cls, celsius, invalid_value=None):
        """Return a Temperature instance intialized with a value in celsius."""
        return cls(celsius, celsius, invalid_value)

    def to_f(self):
        """Return the temperature measurement as farenheit."""
        if self.value is not None:
            return (self.value * 1.8) + 32.0

    @classmethod
    def f_or_c(cls, temperature, measurement_system=DisplayMeasure.metric):
        """Return the temperature measurement as farenheit or celsius."""
        return temperature.c_or_f(measurement_system)

    def c_or_f(self, measurement_system=DisplayMeasure.metric):
        """Return the temperature measurement as farenheit or celsius."""
        return self.value if measurement_system is DisplayMeasure.metric else self.to_f()


class Cadence(Measurement):
    """Object that represents a cadence measurement."""

    def __init__(self, cycles, raw_value, invalid_value):
        """Return a Cadence instance intialized with a value in cycles."""
        super().__init__(cycles, raw_value, invalid_value, self.to_cycles, 'cycles')

    @classmethod
    def from_cycles(cls, cycles, invalid_value=None):
        """Return a Temperature instance intialized with a value in celsius."""
        return cls(cycles, cycles, invalid_value)

    def to_cycles(self):
        """Return the cadence measurement as cycles."""
        return self.value

    def to_spm(self):
        """Return the cadence measurement as steps per minute."""
        if self.value is not None:
            return (self.value * 2)



"""Given a schema, unpack FIT file encoded data into a Python object."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import struct
import logging
import enum


logger = logging.getLogger(__name__)


class Architecture(enum.Enum):
    """The endian type that the FIT file was encoded with."""

    Little_Endian   = 0
    Big_Endian      = 1


class Schema(object):
    """Describes how the data of a FIT file object is encoded."""

    type_to_size = {
        'CHAR': 1,
        'INT8': 1,
        'UINT8': 1,
        'INT16': 2,
        'UINT16': 2,
        'INT32': 4,
        'UINT32': 4,
        'INT64': 8,
        'UINT64': 8,
        'FLOAT32': 4,
        'FLOAT64': 4
    }

    __type_to_unpack_format = {
        'CHAR': 'B',
        'INT8': 'b',
        'UINT8': 'B',
        'INT16': 'h',
        'UINT16': 'H',
        'INT32': 'i',
        'UINT32': 'I',
        'INT64': 'q',
        'UINT64': 'Q',
        'FLOAT32': 'f',
        'FLOAT64': 'd'
    }

    def __init__(self, name, ordered_dict):
        """Return a message schema given it's name and an ordered dict of its fields."""
        self.name = name
        self.ordered_dict = ordered_dict
        self.file_size = [0, 0]
        self.unpack_format = [None, None]

    def __compile_unpack(self, endian):
        if endian is Architecture.Big_Endian:
            unpack_format = '>'
        else:
            unpack_format = ''
        for key in self.ordered_dict:
            (type, count) = self.ordered_dict[key]
            for _ in range(count):
                unpack_format += self.__type_to_unpack_format[type]
                self.file_size[endian.value] += self.type_to_size[type]
        self.unpack_format[endian.value] = unpack_format

    def get_unpack(self, endian):
        """Get a compiled unpack format for this schema."""
        if self.unpack_format[endian.value] is None:
            self.__compile_unpack(endian)
        return (self.unpack_format[endian.value], self.file_size[endian.value])

    def _decode(self, data):
        """Create a dict of message fields given a bytesarray."""
        decoded_data = {}
        index = 0
        for (key, (_, count)) in self.ordered_dict.items():
            if count > 1:
                decoded_data[key] = [data[index + repeat] for repeat in range(count)]
                index += count
            elif count == 1:
                decoded_data[key] = data[index]
                index += 1
        return decoded_data


class Data(object):
    """The base object for decoding FIT file data."""

    def __init__(self, file, primary_schema, secondary_schemas=None, endian=Architecture.Little_Endian):
        """Return a Data instance created by parsing data from a FIT file."""
        self.file = file
        self.primary_schema = primary_schema
        self.secondary_schemas = secondary_schemas
        self.endian = endian
        self.file_size = 0
        self.decode_all()
        self._convert()

    def _decode(self, schema):
        """Given a schema decode file data into fields and add them as properties of the data object."""
        (unpack_format, file_size) = schema.get_unpack(self.endian)
        self.file_size += file_size
        bytes = struct.unpack(unpack_format, self.file.read(file_size))
        vars(self).update(schema._decode(bytes))

    def decode_all(self):
        """Decode file data using all of the data objects schemas."""
        self._decode(self.primary_schema)
        if self.secondary_schemas is not None:
            for schema, control_func in self.secondary_schemas:
                if control_func():
                    self._decode(schema)

    def _convert(self):
        pass



"""Library for parsing FIT files and returning them as lists of messages dictionaries."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .exceptions import FitException, FitFileError, FitFileBadHeaderSize, FitFileBadProtocolVersion, FitFileDataType, FitMessageType, FitMessageParse, FitUndefDevMessageType, FitOutOfOrderMessage



"""Objects defining FIT file message types."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import enum

from .exceptions import FitMessageType


class UnknownMessageType(object):
    """Represents an unknown FIT file message type."""

    def __init__(self, index):
        """Return an instance of UnknownMessageType."""
        self.value = index
        self.name = 'unknown_%d' % index

    def __eq__(self, other):
        return other and self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.value

    def __repr__(self):
        """Return a string representation of the UnknownMessageType instance."""
        return f'<UnknownMessageType.{self.name}: {self.value}>'


class MessageType(enum.Enum):
    """Enum of all known FIT file message types."""

    file_id                         = 0
    capabilities                    = 1
    device_settings                 = 2
    user_profile                    = 3
    hrm_profile                     = 4
    sdm_profile                     = 5
    bike_profile                    = 6
    zones_target                    = 7
    hr_zone                         = 8
    power_zone                      = 9
    met_zone                        = 10
    # 11 is unknown
    sport                           = 12
    # 13,14 are unknown
    unknown_13                      = 13
    unknown_14                      = 14
    goal                            = 15
    # 16,17 are unknown
    session                         = 18
    lap                             = 19
    record                          = 20
    event                           = 21
    # 22 is unknown
    source                          = 22
    device_info                     = 23
    # 24,25 are unknown
    unknown_24                      = 24
    workout                         = 26
    workout_step                    = 27
    schedule                        = 28
    location                        = 29
    weight_scale                    = 30
    course                          = 31
    course_point                    = 32
    totals                          = 33
    activity                        = 34
    software                        = 35
    # 36 not known
    file_capabilities               = 37
    mesg_capabilities               = 38
    field_capabilities              = 39
    # 40-48 not known
    file_creator                    = 49
    blood_pressure                  = 51
    # 52 not known
    speed_zone                      = 53
    # 54 not known
    monitoring                      = 55
    # 56-71 not known
    training_file                   = 72
    # 73-77 not known
    hrv                             = 78
    # 79 not known
    ant_rx                          = 80
    ant_tx                          = 81
    # 83-100 not known
    ant_channel_id                  = 82
    length                          = 101
    # 102 not known
    monitoring_info                 = 103
    battery                         = 104
    pad                             = 105
    slave_device                    = 106
    personal_record                 = 114
    # 107-126 not known
    connectivity                    = 127
    weather_conditions              = 128
    weather_alert                   = 129
    # 130 not known
    cadence_zone                    = 131
    hr                              = 132
    # 133-141 not known
    firstbeat                       = 140
    unknown_141                     = 141
    segment_lap                     = 142
    # 143, 144 not known
    memo_glob                       = 145
    # 146 not known
    sensor                          = 147
    segment_id                      = 148
    segment_leaderboard_entry       = 149
    segment_point                   = 150
    segment_file                    = 151
    # 152-157 not known
    workout_session                 = 158
    watchface_settings              = 159
    gps_metadata                    = 160
    camera_event                    = 161
    timestamp_correlation           = 162
    # 163 not known
    gyroscope_data                  = 164
    accelerometer_data              = 165
    # 166 not known
    three_d_sensor_calibration      = 167
    # 168 not known
    video_frame                     = 169
    # 170-173 not known
    obdii_data                      = 174
    # 175,176 not known
    nmea_sentence                   = 177
    aviation_attitude               = 178
    # 179-183 not known
    video                           = 184
    video_title                     = 185
    video_description               = 186
    video_clip                      = 187
    ohr_settings                    = 188
    # 189-199 not known
    exd_screen_configuration        = 200
    exd_data_field_configuration    = 201
    exd_data_concept_configuration  = 202
    # 203-205 not known
    field_description               = 206
    dev_data_id                     = 207
    magnetometer_data               = 208
    barometer_data                  = 209
    one_d_sensor_calibration        = 210
    # 211-215 not known
    hr_zones_timer                  = 216
    # 217-224 not known
    set                             = 225
    # 226 not known
    stress_level                    = 227
    # 228-241 not known
    unknown_233                     = 233
    local_time                      = 241
    # 242-257 not known
    dive_settings                   = 258
    dive_gas                        = 259
    # 260,261 not known
    dive_alarm                      = 262
    # 263 not known
    exercise_title                  = 264
    # 265-267 not known
    dive_summary                    = 268
    pulse_ox                        = 269
    start                           = 273   # this is a guess
    sleep_data                      = 274   # this is a guess
    sleep_level                     = 275   # this is a guess
    end                             = 276   # this is a guess
    unknown_284                     = 284
    jump                            = 285
    respiration                     = 297
    #
    climb_pro                       = 317
    #
    #
    #
    mfg_range_min                   = 0xFF00
    mfg_range_max                   = 0xFFFE

    def is_unknown(self):
        """Return if the message type is not a known type."""
        return "unknown" in self.name

    @classmethod
    def get_type(cls, message_number):
        """Given a message number, return the message type."""
        if message_number < 0 or message_number > cls.mfg_range_max.value:
            raise FitMessageType(f'Message number {message_number} out of range')
        try:
            return cls(message_number)
        except ValueError:
            return UnknownMessageType(message_number)



"""Enums that represent FIT file message sport and sub-sport field values."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

from .field_enums import FieldEnum


class Sport(FieldEnum):
    """Enums that represent FIT file message sport field values."""

    generic                 = 0
    running                 = 1
    cycling                 = 2
    transition              = 3
    fitness_equipment       = 4
    swimming                = 5
    basketball              = 6
    soccer                  = 7
    tennis                  = 8
    american_football       = 9
    training                = 10
    walking                 = 11
    cross_country_skiing    = 12
    alpine_skiing           = 13
    snowboarding            = 14
    rowing                  = 15
    mountaineering          = 16
    hiking                  = 17
    multisport              = 18
    paddling                = 19
    flying                  = 20
    e_biking                = 21
    motorcycling            = 22
    boating                 = 23
    driving                 = 24
    golf                    = 25
    hang_gliding            = 26
    horseback_riding        = 27
    hunting                 = 28
    fishing                 = 29
    inline_skating          = 30
    rock_climbing           = 31
    sailing                 = 32
    ice_skating             = 33
    sky_diving              = 34
    snowshoeing             = 35
    snowmobiling            = 36
    stand_up_paddleboarding = 37
    surfing                 = 38
    wakeboarding            = 39
    water_skiing            = 40
    kayaking                = 41
    rafting                 = 42
    windsurfing             = 43
    kitesurfing             = 44
    tactical                = 45
    jumpmaster              = 46
    boxing                  = 47
    floor_climbing          = 48
    all                     = 254
    invalid                 = 255

    def preferred(self):
        """Return if this Sport enum instance is a preferred value or not."""
        return self.value > Sport.generic.value and self.value < Sport.all.value


class SubSport(FieldEnum):
    """An enum field containing more specific sport information."""

    generic                 = 0
    treadmill               = 1
    street                  = 2
    trail                   = 3
    track                   = 4
    spin                    = 5
    indoor_cycling          = 6
    road                    = 7
    mountain                = 8
    downhill                = 9
    recumbent               = 10
    cyclocross              = 11
    hand_cycling            = 12
    track_cycling           = 13
    indoor_rowing           = 14
    elliptical              = 15
    stair_climbing          = 16
    lap_swimming            = 17
    open_water              = 18
    flexibility_training    = 19
    strength_training       = 20
    warm_up                 = 21
    match                   = 22
    exercise                = 23
    challenge               = 24
    indoor_skiing           = 25
    cardio_training         = 26
    indoor_walking          = 27
    e_bike_fitness          = 28
    bmx                     = 29
    casual_walking          = 30
    speed_walking           = 31
    bike_to_run_transition  = 32
    run_to_bike_transition  = 33
    swim_to_bike_transition = 34
    atv                     = 35
    motocross               = 36
    backcountry             = 37
    resort                  = 38
    rc_drone                = 39
    wingsuit                = 40
    whitewater              = 41
    skate_skiing            = 42
    yoga                    = 43
    pilates                 = 44
    indoor_running          = 45
    gravel_cycling          = 46
    e_bike_mountain         = 47
    commuting               = 48
    mixed_surface           = 49
    navigate                = 50
    track_me                = 51
    map                     = 52
    single_gas_diving       = 53
    multi_gas_diving        = 54
    guage_diving            = 55
    apnea_diving            = 56
    apneas_hunting          = 57
    virtual_activity        = 58
    obstacle                = 59
    all                     = 254
    invalid                 = 255

    def preferred(self):
        """Return if this SubSport enum instance is a preferred value or not."""
        return self.value > SubSport.generic.value and self.value < SubSport.all.value



"""Code that manages a FIT file."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import logging
import datetime

from .file_header import FileHeader
from .record_header import RecordHeader, MessageClass
from .definition_message import DefinitionMessage
from .data_message import DataMessageDecodeContext, DataMessage
from .message_type import MessageType
from .field_enums import DisplayMeasure


logger = logging.getLogger(__name__)

name_regex = r'\w+\.(fit|FIT)'


# dynamically generated class properties
# pylint: disable=no-member
class File(object):
    """Object that represents a FIT file."""

    def __init__(self, filename, measurement_system=DisplayMeasure.metric):
        """
        Return a File instance by parsing a FIT file.

        Parameters:
        ----------
            filename (string): The name of the FIT file including full path.
            measurement_system (DisplayMeasure): The measurement units (metric, statute, etc) to uwe when parsing the FIT file.

        """
        self.filename = filename
        self.measurement_system = measurement_system
        self.message_types = []
        self.messages = []
        for message_type in MessageType:
            vars(self)[message_type.name] = []
        with open(filename, 'rb') as file:
            self.__parse(file)
        self.__sumarize()

    def __parse(self, file):
        logger.debug("Parsing File %s", self.filename)
        self.file_header = FileHeader(file)
        self.data_size = self.file_header.data_size
        self._definition_messages = {}
        self.__dev_fields = {}
        data_consumed = 0
        self.record_count = 0
        data_message_context = DataMessageDecodeContext()
        while self.data_size > data_consumed:
            record_header = RecordHeader(file)
            local_message_num = record_header.local_message()
            data_consumed += record_header.file_size
            self.record_count += 1
            logger.debug("Parsed record %r", record_header)
            if record_header.message_class is MessageClass.definition:
                definition_message = DefinitionMessage(record_header, self.__dev_fields, file)
                logger.debug("  Definition [%d]: %s", local_message_num, definition_message)
                data_consumed += definition_message.file_size
                self._definition_messages[local_message_num] = definition_message
            else:
                definition_message = self._definition_messages[local_message_num]
                data_message = DataMessage(definition_message, file, self.measurement_system, data_message_context)
                logger.debug("  Data [%d]: %s", local_message_num, data_message)
                data_consumed += data_message.file_size
                data_message_type = data_message.type
                if data_message_type == MessageType.field_description:
                    self.__dev_fields[data_message.fields.field_definition_number] = data_message
                logger.debug("Parsed %r", data_message_type)
                self.__save_message(data_message_type, data_message)
            logger.debug("Record %d: consumed %d of %s %r", self.record_count, data_consumed, self.data_size, self.measurement_system)
        self.last_message_timestamp = data_message_context.last_timestamp

    def __save_message(self, data_message_type, data_message):
        if data_message_type.name in vars(self):
            vars(self)[data_message_type.name].append(data_message)
        else:
            vars(self)[data_message_type.name] = [data_message]
        self.messages.append(data_message)
        if data_message_type not in self.message_types:
            self.message_types.append(data_message_type)

    def __sumarize(self):
        first_file_id = self.file_id[0]
        self.time_created = first_file_id.fields.time_created
        self.type = first_file_id.fields.type
        self.product = first_file_id.fields.product
        self.serial_number = first_file_id.fields.serial_number
        self.device = f'{self.product}_{self.serial_number}'
        if MessageType.device_settings in self.message_types:
            self.utc_offset = self.device_settings[0].fields.time_offset
        elif MessageType.monitoring_info in self.message_types:
            first_monitoring_info = self.monitoring_info[0]
            monitoring_info_time_utc = first_monitoring_info.fields.timestamp
            monitoring_info_time_local = first_monitoring_info.fields.local_timestamp
            self.utc_offset = (monitoring_info_time_local - monitoring_info_time_utc.replace(tzinfo=None)).total_seconds()
        else:
            self.utc_offset = 0
        self.local_tz = datetime.timezone(datetime.timedelta(seconds=self.utc_offset))
        self.time_created_local = self.utc_datetime_to_local(self.time_created)
        if self.last_message_timestamp is not None:
            self.time_ended_local = self.utc_datetime_to_local(self.last_message_timestamp)
        else:
            self.time_ended_local = self.time_created_local
        if MessageType.sport in self.message_types:
            self.sport_type = self.sport[0].fields.sport
            self.sub_sport_type = self.sport[0].fields.sub_sport
        else:
            self.sport_type = None
            self.sub_sport_type = None
        if MessageType.dev_data_id in self.message_types:
            self.dev_application_id = self.dev_data_id[0].fields.application_id
        else:
            self.dev_application_id = None
        self.dev_fields = {msg.fields.field_name: {'native_message_num': msg.fields.native_message_num, 'units': msg.fields.units}for msg in self.field_description}

    def date_span(self):
        """Return a tuple of the start and end dates of the file."""
        return (self.time_created, self.last_message_timestamp)

    def utc_datetime_to_local(self, dt):
        """Return a local datetime based on the passed in UTC datetime and the file's known UTC offset."""
        if self.local_tz is not None and dt.tzinfo is datetime.timezone.utc:
            return dt.astimezone(self.local_tz).replace(tzinfo=None)
        return dt.replace(tzinfo=None)

    def __getitem__(self, message_type):
        """Return the attribute named name."""
        return vars(self).get(message_type.name, [])

    def __str__(self):
        """Return a string representation of the class instance."""
        return f'File({repr(self.type)} {self.filename} {self.type} {repr(self.message_types)} dev fields {repr(self.dev_fields)})'



"""Library for parsing FIT files and returning them as lists of messages dictionaries."""

import re
import os
from setuptools import setup


def get_version(version_file):
    """Extract version fron module source."""
    with open(version_file, 'r') as file:
        data = file.read()
        match = re.search(r'version_info = \((\d), (\d), (\d)\)', data, re.M)
        if match:
            return f'{match.group(1)}.{match.group(2)}.{match.group(3)}'


def get_long_description(readme_file):
    """Extract long description fron the module readme."""
    with open(readme_file, "r", encoding="utf-8") as file:
        return file.read()


module_name = 'fitfile'
module_version = get_version(module_name + os.sep + 'version_info.py')
module_long_description = get_long_description('README.rst')

print(f"Building {module_name} {module_version}\n{module_long_description}")

setup(name=module_name, version=module_version, author='Tom Goetz', packages=[module_name, f'{module_name}.conversions', f'{module_name}.exceptions', f'{module_name}.field_enums'],
      description='Decode FIT format files.',
      long_description=module_long_description,
      long_description_content_type='text/x-rst',
      url="https://github.com/tcgoetz/Fit",
      classifiers=[
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent"
      ],
      python_requires=">=3.0")



"""Objects that represent FIT file user message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .enum_fields import EnumField
from .field_enums import Gender, Language


class GenderField(EnumField):
    """A FIT file message field containing the user's gender."""

    _name = 'gender'
    _enum = Gender


class LanguageField(EnumField):
    """A FIT file message field containing the user's language."""

    _name = 'language'
    _enum = Language



"""Object that represents a FIT file developer data field."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import collections

from .data import Data, Schema


class DevDataField(Data):
    """Object that represents a FIT file developer data field."""

    def __init__(self, file, definition_message, dev_field_definition, measurement_system):
        """Return a new DevDataField instance created by reading the FIT file and applying the supplied definitions."""
        self.dev_field_definition = dev_field_definition
        self.measurement_system = measurement_system
        self.field = dev_field_definition.field()
        type = dev_field_definition.type_string()
        count = dev_field_definition.type_count()
        self.field_value = None
        schema = Schema('DevDataField', collections.OrderedDict([('field_value', [type, count])]))
        super().__init__(file, schema, None, definition_message.endian)

    def _convert(self):
        self.value = self.field.convert(self.field_value, self.dev_field_definition.invalid(), self.measurement_system)[0]

    def __str__(self):
        """Return a string representation of the DevDataField instance."""
        return str(self.value)



"""Objects that represent FIT file object message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import Field
from .field_enums import DisplayMeasure
from .field_value import FieldValue
from .measurement import Distance, Speed, Weight, Longitude, Latitude, Temperature


class ObjectField(Field):
    """Class that handles a field that translates into a Python object."""

    def __init__(self, obj_func, output_func, **kwargs):
        """Return an instance of ObjectField."""
        self.obj_func = obj_func
        self.output_func = output_func
        super().__init__(**kwargs)

    def _invalid_single(self, value, invalid):
        return value.is_invalid()

    def _convert_single(self, value, invalid):
        return self.output_func(value, self.measurement_system)

    def convert(self, value, invalid, measurement_system=DisplayMeasure.metric):
        """Return a FieldValue containing the field value as a Python object."""
        self.measurement_system = measurement_system
        # apply scle and offset to invalid to or tests for invalid will fail!!
        value_obj = self.obj_func((value / self._scale) - self._offset, (invalid / self._scale) - self._offset)
        return [FieldValue(self, value_obj, invalid, **{self._name: self._convert_many(value_obj, invalid)})]

    def reconvert(self, value_obj, invalid, measurement_system=DisplayMeasure.metric):
        """Return a FieldValue containing the field value as a Python object."""
        self.measurement_system = measurement_system
        return {self._name: self._convert_many(value_obj, invalid)}


class HeightField(ObjectField):
    """Class that handles a user height measurement from a FIT message field."""

    _name = 'height'

    def __init__(self):
        """Return a HeightField instance."""
        super().__init__(Distance.from_cm, Distance.feet_or_meters)


class WeightField(ObjectField):
    """Class that handles a user weight measurement from a FIT message field."""

    def __init__(self, name):
        """Return a WeightField instance."""
        super().__init__(Weight.from_cgs, Weight.lbs_or_kgs, name=name)


class SpeedMpsField(ObjectField):
    """Field holding a distance measure in meters per second."""

    def __init__(self, name):
        """Return a SpeedMpsField instance."""
        super().__init__(Speed.from_mmps, Speed.mph_or_kph, name=name)


class LongitudeField(ObjectField):
    """Field that handles a longitude measure in semicircles."""

    def __init__(self, name):
        """Return a LongitudeField instance."""
        super().__init__(Longitude.from_semicircles, Longitude.to_degrees, name=name)


class LatiitudeField(ObjectField):
    """Field that handles a latitude measure in semicircles."""

    def __init__(self, name):
        """Return a LatiitudeField instance."""
        super().__init__(Latitude.from_semicircles, Latitude.to_degrees, name=name)


class TemperatureField(ObjectField):
    """Field holding a temperature measurement in celsius."""

    def __init__(self, name):
        """Return a ObjectField instance."""
        super().__init__(Temperature.from_celsius, Temperature.f_or_c, name=name)


class DistanceMetersField(ObjectField):
    """Field holding a distance measurement in meters."""

    def __init__(self, name, obj_func=Distance.from_meters, output_func=Distance.feet_or_meters, **kwargs):
        """Return a new instance of DistanceMetersField."""
        super().__init__(obj_func, output_func, name=name, **kwargs)


class EnhancedDistanceMetersField(DistanceMetersField):
    """Field holding a distance measure in meters."""

    def __init__(self, name):
        """Return a EnhancedDistanceMetersField instance."""
        super().__init__(name, Distance.from_mm, Distance.feet_or_meters)


class DistanceCentimetersToKmsField(DistanceMetersField):
    """Field holding a distance measure in meters."""

    def __init__(self, name):
        """Return a DistanceCentimetersToKmsField instance."""
        super().__init__(name, Distance.from_cm, Distance.kms_or_miles)


class DistanceCentimetersToMetersField(DistanceMetersField):
    """Field holding a distance measure in meters."""

    def __init__(self, name):
        """Return a DistanceCentimetersToMetersField instance."""
        super().__init__(name, Distance.from_cm, Distance.feet_or_meters)


class DistanceMillimetersToMetersField(DistanceMetersField):
    """Field holding a distance measure in meters."""

    def __init__(self, name):
        """Return a DistanceMillimetersToMetersField instance."""
        super().__init__(name, Distance.from_mm, Distance.feet_or_meters)


class DistanceMillimetersField(DistanceMetersField):
    """Field holding a distance measure in meters."""

    def __init__(self, name):
        """Return a DistanceMillimetersField instance."""
        super().__init__(name, Distance.from_mm, Distance.inches_or_mm, scale=10.0)


class AltitudeField(DistanceMetersField):
    """A field containing a altitude reading with greater range."""

    def __init__(self, name):
        """Return an instance of AltitudeField."""
        super().__init__(name, Distance.from_meters, Distance.feet_or_meters, scale=5.0, offset=500.0)


class CompressedSpeedDistanceField(Field):
    """A field that generates sub fields fields for activity and intensity."""

    _name = 'compressed_speed_distance'

    speed_field = SpeedMpsField('speed')
    distance_field = DistanceCentimetersToKmsField('distance')

    def convert(self, value, invalid, measurement_system):
        """Convert the value to sub fields."""
        self.measurement_system = measurement_system
        speed = value & 0x3f
        distance = value >> 12
        return self.speed_field.convert(speed, 0xff, measurement_system) + self.distance_field.convert(distance, 0xff, measurement_system)
